/*
 * SPDX-FileCopyrightText: 2025 Florian Müllner <fmuellner@gnome.org>
 * SPDX-License-Identifier: LGPL-2.1-or-later
 */
#pragma once

#include <shew-external-window.h>
#include <shew-window-exporter.h>
